#include <rclcpp/rclcpp.hpp>
#include <cv_bridge/cv_bridge.h>
#include <geometry_msgs/msg/polygon_stamped.hpp>
#include <sensor_msgs/msg/image.hpp>

/*** 名前空間を省略して利用できるように宣言 ***/
using geometry_msgs::msg::Point32;
using geometry_msgs::msg::PolygonStamped;
using sensor_msgs::msg::Image;

namespace practice_scamper
{

class ColorExtractor : public rclcpp::Node
{
public:
  ColorExtractor(rclcpp::NodeOptions options = rclcpp::NodeOptions());
  ~ColorExtractor(){}
private:
  void initParameter();
  void initTopic();
  void extractColor(const cv::Mat &src, cv::Mat &dst);
  void labeling(const cv::Mat &src, cv::Rect &rect);
  void rect2Points(cv::Size Image_size, cv::Rect rect, std::vector<Point32> &points);
  /*** コールバック関数 ***/
  void onImageSubscribed(const Image::SharedPtr img);
  /*** メンバー変数 ***/
  rclcpp::Publisher<PolygonStamped>::SharedPtr pub_rect_;
  rclcpp::Subscription<Image>::SharedPtr sub_img_;
  /*** パラメータ ***/
  int hue_min_, hue_max_; // 色相の閾
  int sat_min_, sat_max_; // 彩度の閾値
  int val_min_, val_max_; // 明度の閾値
  int rect_size_min_; // 矩形の大きさの閾値
};
/*** コンストラクタ ***/
ColorExtractor::ColorExtractor(rclcpp::NodeOptions options) : Node("color_extractor", options)
{
  initParameter();
  initTopic();
}
/*** パラメータの読み込み ***/
void ColorExtractor::initParameter()
{
  hue_min_ = this->declare_parameter<int>("hue_min", 10);
  hue_max_ = this->declare_parameter<int>("hue_max", 30);
  sat_min_ = this->declare_parameter<int>("saturation_min", 150);
  sat_max_ = this->declare_parameter<int>("saturation_max", 255);
  val_min_ = this->declare_parameter<int>("value_min", 150);
  val_max_ = this->declare_parameter<int>("value_max", 255);
  rect_size_min_ = this->declare_parameter<int>("rect_size_min", 250);
}
/*** トピックの初期化 ***/
void ColorExtractor::initTopic()
{
  using std::placeholders::_1;
  /*** パブリッシャの初期化 ***/
  pub_rect_ = this->create_publisher<PolygonStamped>("rectangle", 10);
  /*** サブスクライバの初期化 ***/
  sub_img_ = this->create_subscription<Image>("/scamper_eyes/left/image", 10, 
                                              std::bind(&ColorExtractor::onImageSubscribed, this, _1));
}
/*** 特定色の抽出 ***/
void ColorExtractor::extractColor(const cv::Mat &src, cv::Mat &dst)
{
  /*** 入力画像をHSV表色系に変換 ***/
  cv::Mat hsv;
  cv::cvtColor(src, hsv, cv::COLOR_BGR2HSV);
  /*** 出力画像を0で初期化 ***/
  dst = cv::Mat(src.size(), CV_8UC1, cv::Scalar(0));
  /*** 色相(H)，彩度(S)が閾値内の画素のみ255を代入 ***/
  for(int y = 0; y < hsv.rows; y++){
    for(int x = 0; x < hsv.cols; x++){
      cv::Vec3b val = hsv.at<cv::Vec3b>(y, x);
      if(val[0] < hue_min_ || val[0] > hue_max_) continue;
      if(val[1] < sat_min_ || val[1] > sat_max_) continue;
      if(val[2] < val_min_ || val[2] > val_max_) continue;
      dst.at<uchar>(y, x) = 255;
    }
  }
  /*** メディアンフィルタを適用しノイズ成分を除去 ***/
  cv::medianBlur(dst, dst, 5);
}
/*** ラベリング処理 ***/
void ColorExtractor::labeling(const cv::Mat &src, cv::Rect &rect)
{
  cv::Mat labels, stats, centroids;
  int n = cv::connectedComponentsWithStats(src, labels, stats, centroids, 8);
  /*** ラベリング領域が最大のものを選択 ***/
  rect = cv::Rect(0, 0, 0, 0);
  for(int i = 1; i < n; i++){
    int *stat = stats.ptr<int>(i);
    int size = stat[cv::ConnectedComponentsTypes::CC_STAT_AREA];
    if(size < rect_size_min_) continue;
    if(rect.width * rect.height < size){
      rect.x = stat[cv::ConnectedComponentsTypes::CC_STAT_LEFT];
      rect.y = stat[cv::ConnectedComponentsTypes::CC_STAT_TOP];
      rect.width = stat[cv::ConnectedComponentsTypes::CC_STAT_WIDTH];
      rect.height = stat[cv::ConnectedComponentsTypes::CC_STAT_HEIGHT];
    }
  }
}
/*** cv::Rect -> Polygonへの変換 ***/
void ColorExtractor::rect2Points(cv::Size image_size, cv::Rect rect,std::vector<Point32> &points)
{
  points.resize(4);
  points[0].x = image_size.height / 2 - rect.y;
  points[0].y = rect.x - image_size.width / 2;
  points[1].x = image_size.height / 2 - (rect.y + rect.height);
  points[1].y = rect.x - image_size.width / 2;
  points[2].x = image_size.height / 2 - (rect.y + rect.height);
  points[2].y = (rect.x + rect.width) - image_size.width / 2;
  points[3].x = image_size.height / 2 - rect.y;
  points[3].y = (rect.x + rect.width) - image_size.width / 2;
}
/*** カメラから画像を受信した時に呼び出されるコールバック関数 ***/
void ColorExtractor::onImageSubscribed(const Image::SharedPtr img)
{
  auto cv_img = cv_bridge::toCvShare(img);
  cv::Mat binary_img;
  cv::Rect rect;
  auto polygon = std::make_unique<PolygonStamped>();
  /*** 特定色の抽出 ***/
  extractColor(cv_img->image, binary_img);
  /*** ラベリング ***/
  labeling(binary_img, rect);
  /*** cv::Rect -> Polygonへの変換 ***/
  rect2Points(cv_img -> image.size(),rect, polygon->polygon.points);
  polygon->header.stamp = img->header.stamp;
  /*** 矩形情報のパブリッシュ ***/
  pub_rect_->publish(std::move(polygon));
}

}
/*** ColorExtractorクラスをコンポーネントとして登録 ***/
#include <rclcpp_components/register_node_macro.hpp>
RCLCPP_COMPONENTS_REGISTER_NODE(practice_scamper::ColorExtractor)
