#include <rclcpp/rclcpp.hpp>
#include <cv_bridge/cv_bridge.h>
#include <message_filters/subscriber.h>
#include <message_filters/sync_policies/approximate_time.h>
#include <geometry_msgs/msg/polygon_stamped.hpp>
#include <sensor_msgs/msg/image.hpp>
#include <opencv2/opencv.hpp>

/*** 名前空間を省略して利用できるように宣言 ***/
using geometry_msgs::msg::PolygonStamped;
using sensor_msgs::msg::Image;
using message_filters::Subscriber;
using message_filters::sync_policies::ApproximateTime;
using message_filters::Synchronizer;
/*** typedefを用いた型の再定義 ***/
typedef ApproximateTime<Image, PolygonStamped> SyncPolicy;

namespace practice_scamper
{

class ColorExtractorViewer : public rclcpp::Node
{
public:
  ColorExtractorViewer(rclcpp::NodeOptions options = rclcpp::NodeOptions());
  ~ColorExtractorViewer(){}
private:
  void initTopic();
  /*** コールバック関数 ***/
  void onResultSubscribed(Image::ConstSharedPtr img,PolygonStamped::ConstSharedPtr polygon);
  /*** メンバー変数 ***/
  std::shared_ptr<Subscriber<Image>> sub_img_;
  std::shared_ptr<Subscriber<PolygonStamped>> sub_polygon_;
  std::shared_ptr<Synchronizer<SyncPolicy>> sync_msgs_;
};
/*** コンストラクタ ***/
ColorExtractorViewer::ColorExtractorViewer(rclcpp::NodeOptions options) : Node("color_extractor_viewer", options)
{
  initTopic();
}
/*** トピックの初期化 ***/
void ColorExtractorViewer::initTopic()
{
  using std::placeholders::_1;
  using std::placeholders::_2;
  /*** サブスクライバの初期化 ***/
  sub_img_.reset(new Subscriber<Image>(this, "/scamper_eyes/left/image"));
  sub_polygon_.reset(new Subscriber<PolygonStamped>(this, "rectangle"));
  sync_msgs_.reset(new Synchronizer<SyncPolicy>(SyncPolicy(10), *sub_img_, *sub_polygon_));
  sync_msgs_->registerCallback(std::bind(&ColorExtractorViewer::onResultSubscribed, this, _1, _2));
}
/*** 画像と矩形領域を受信した時に呼び出されるコールバック関数 ***/
void ColorExtractorViewer::onResultSubscribed(Image::ConstSharedPtr img, PolygonStamped::ConstSharedPtr polygon)
{
  auto cv_img = cv_bridge::toCvCopy(img);
  /*** Polygon -> cv::Rectの変換 ***/
  cv::Point p1(polygon->polygon.points[0].y + img->width / 2, img->height / 2 - polygon->polygon.points[0].x);
  cv::Point p2(polygon->polygon.points[2].y + img->width / 2, img->height / 2 - polygon->polygon.points[2].x);
  cv::Rect rect(p1, p2);
  /*** 画像に矩形領域を描画 ***/
  cv::rectangle(cv_img->image, rect, cv::Scalar(0, 0, 255), 2);
  /*** 描画ウィンドウの表示 ***/
  cv::namedWindow("ColorExtractorViewer");
  cv::imshow("ColorExtractorViewer", cv_img->image);
  cv::waitKey(1);
}
}
/*** ColorExtractorViewerクラスをコンポーネントとして登録 ***/
#include <rclcpp_components/register_node_macro.hpp>
RCLCPP_COMPONENTS_REGISTER_NODE(practice_scamper::ColorExtractorViewer)