#include <rclcpp/rclcpp.hpp>
#include <geometry_msgs/msg/polygon_stamped.hpp>
#include <geometry_msgs/msg/twist.hpp>

/*** 名前空間を省略して利用できるように宣言 ***/
using geometry_msgs::msg::Point32;
using geometry_msgs::msg::PolygonStamped;
using geometry_msgs::msg::Twist;

namespace practice_scamper
{

class FollowColor : public rclcpp::Node
{
public:
  FollowColor(rclcpp::NodeOptions options = rclcpp::NodeOptions());
  ~FollowColor(){}
private:
  void initParameter();
  void initTopic();
  Twist points2Twist(const std::vector<Point32> &points);
  /*** コールバック関数 ***/
  void onPolygonSubscribed(const PolygonStamped::SharedPtr polygon);
  /*** メンバー変数 ***/
  rclcpp::Publisher<Twist>::SharedPtr pub_vel_;
  rclcpp::Subscription<PolygonStamped>::SharedPtr sub_polygon_;
  /*** パラメータ ***/
  double max_vel_; // ロボットの最大速度
  double vel_coefficient_; // 速度計算のための係数
  int dead_zone_; // 物体が正面に存在すると判断される矩形中心の閾値
};
/*** コンストラクタ ***/
FollowColor::FollowColor(rclcpp::NodeOptions options) : Node("follow_color", options)
{
  initParameter();
  initTopic();
}
/*** パラメータの読み込み ***/
void FollowColor::initParameter()
{
  max_vel_ = this->declare_parameter<double>("max_velocity", 1.0);
  vel_coefficient_ = this->declare_parameter<double>("velocity_coefficient", -0.01);
  dead_zone_ = this->declare_parameter<int>("dead_zone", 10);
}
/*** トピックの初期化 ***/
void FollowColor::initTopic()
{
  using std::placeholders::_1;
  /*** パブリッシャの初期化 ***/
  pub_vel_ = this->create_publisher<Twist>("/scamper_driver/cmd_robot_vel", 10);
  /*** サブスクライバの初期化 ***/
  sub_polygon_ = this->create_subscription<PolygonStamped>("rectangle", 10, 
                                                           std::bind(&FollowColor::onPolygonSubscribed, this, _1));
}
/*** 矩形領域からロボット速度へ変換するための関数 ***/
Twist FollowColor::points2Twist(const std::vector<Point32> &points)
{
  Twist vel;
  /*** 矩形が検出されていなければ停止 ***/
  if(points[0].y == points[2].y) return vel;
  /*** 矩形の中心を計算 ***/
  int c = (points[0].y + points[2].y) / 2;
  /*** 矩形の中心位置がdead_zone_より大きければ回転速度を計算 ***/
  if(abs(c) > dead_zone_) vel.angular.z = c * vel_coefficient_;
  if(vel.angular.z > max_vel_) vel.angular.z = max_vel_;
  else if(vel.angular.z < -max_vel_) vel.angular.z = -max_vel_;
  return vel;
}
/*** カメラから画像を受信した時に呼び出されるコールバック関数 ***/
void FollowColor::onPolygonSubscribed(const PolygonStamped::SharedPtr polygon)
{
  /*** Polygon -> Twistの変換 ***/
  auto vel = std::make_unique<Twist>();
  *vel = points2Twist(polygon->polygon.points);
  /*** ロボット速度のパブリッシュ ***/
  pub_vel_->publish(std::move(vel));
}

}
/*** FollowColorクラスをコンポーネントとして登録 ***/
#include <rclcpp_components/register_node_macro.hpp>
RCLCPP_COMPONENTS_REGISTER_NODE(practice_scamper::FollowColor)