#include <rclcpp/rclcpp.hpp>
#include <geometry_msgs/msg/point_stamped.hpp>
#include <ros_dobot_interfaces/srv/set_ptp_command.hpp>
#include <std_srvs/srv/empty.hpp>
#include <tf2_ros/transform_listener.h>
#include <tf2_geometry_msgs/tf2_geometry_msgs.h>

/*** 名前空間を省略して利用できるように宣言 ***/
using geometry_msgs::msg::Point;
using geometry_msgs::msg::PointStamped;
using geometry_msgs::msg::TransformStamped;
using namespace ros_dobot_interfaces::srv;
using namespace std_srvs::srv;

namespace practice_scamper
{

class GraspBall : public rclcpp::Node
{
public:
  GraspBall(rclcpp::NodeOptions options = rclcpp::NodeOptions());
  ~GraspBall(){}
private:
  void initParameter();
  void initTopic();
  void initService();
  void grasp(PointStamped point);
  void reset();
  /*** コールバック関数 ***/
  void onPointSubscribed(const PointStamped::SharedPtr point);
  void serviceReset(const std::shared_ptr<rmw_request_id_t> header,
                    const std::shared_ptr<Empty::Request> req,
                    const std::shared_ptr<Empty::Response> res);
  /*** メンバー変数 ***/
  rclcpp::Subscription<PointStamped>::SharedPtr sub_point_;
  rclcpp::Client<SetPTPCommand>::SharedPtr cli_move_;
  rclcpp::Client<Empty>::SharedPtr cli_hold_, cli_release_, cli_free_;
  rclcpp::Client<Empty>::SharedPtr cli_go_home_;
  rclcpp::Service<Empty>::SharedPtr srv_reset_;
  rclcpp::Clock::SharedPtr clock_;
  std::shared_ptr<tf2_ros::Buffer> tf_buffer_;
  std::shared_ptr<tf2_ros::TransformListener> tf_listener_;
  bool is_reset_ = false, is_service_done_ = false;
  /*** パラメータ ***/
  std::string camera_name_, manipulator_name_;
  double x_min_, y_min_, z_min_;
  double x_max_, y_max_, z_max_;
};
/*** コンストラクタ ***/
GraspBall::GraspBall(rclcpp::NodeOptions options) : Node("grasp_ball", options)
{
  initParameter();
  initTopic();
  initService();
  clock_ = std::make_shared<rclcpp::Clock>(RCL_ROS_TIME);
  tf_buffer_ = std::make_shared<tf2_ros::Buffer>(clock_);
  tf_listener_ = std::make_shared<tf2_ros::TransformListener>(*tf_buffer_);
}
/*** パラメータの読み込み ***/
void GraspBall::initParameter()
{
  camera_name_ = declare_parameter<std::string>("camera_name", "camera");
  manipulator_name_ = declare_parameter<std::string>("manipulator_name", "dobot");
  x_min_ = declare_parameter<double>("range.min.x", 0.05);
  x_max_ = declare_parameter<double>("range.max.x", 0.35);
  y_min_ = declare_parameter<double>("range.min.y", -0.20);
  y_max_ = declare_parameter<double>("range.max.y", 0.20);
  z_min_ = declare_parameter<double>("range.min.z", -0.10);
  z_max_ = declare_parameter<double>("range.max.z", 0.10);
}
/*** トピックの初期化 ***/
void GraspBall::initTopic()
{
  using std::placeholders::_1;
  /*** サブスクライバの初期化 ***/
  sub_point_ = this->create_subscription<PointStamped>("point", 10, std::bind(&GraspBall::onPointSubscribed, this, _1));
}
/*** サービスの初期化 ***/
void GraspBall::initService()
{
  using namespace std::placeholders;
  cli_move_ = create_client<SetPTPCommand>("/dobot_driver/set_ptp_command");
  cli_hold_ = create_client<Empty>("/dobot_driver/set_gripper_hold");
  cli_release_ = create_client<Empty>("/dobot_driver/set_gripper_release");
  cli_free_ = create_client<Empty>("/dobot_driver/set_gripper_free");
  cli_go_home_ = create_client<Empty>("/dobot_driver/go_to_home");
  srv_reset_ = create_service<Empty>("reset", std::bind(&GraspBall::serviceReset, this, _1, _2, _3));
}
/*** ロボットアームで目標物を掴むための関数 ***/
void GraspBall::grasp(PointStamped point)
{
  /*** ロボットアームのグリッパーを開放 ***/
  auto req_release = std::make_shared<Empty::Request>();
  auto res_release = cli_release_->async_send_request(req_release);
  res_release.wait();
  /*** ロボットアームのグリッパーを指定位置に移動 ***/
  auto req_move = std::make_shared<SetPTPCommand::Request>();
  req_move->x = point.point.x;
  req_move->y = point.point.y;
  req_move->z = point.point.z;
  req_move->r = 0;
  req_move->mode = 1;
  auto res_move = cli_move_->async_send_request(req_move);
  res_move.wait();
  /*** グリッパーを閉じる ***/
  auto req_hold = std::make_shared<Empty::Request>();
  auto res_hold = cli_hold_->async_send_request(req_hold);
  res_hold.wait();
  /*** アームをホームポジションに移動 ***/
  auto req_home = std::make_shared<Empty::Request>();
  auto res_home = cli_go_home_->async_send_request(req_home);
  res_home.wait();
}
/*** システムをリセットするサービスの実体 ***/
void GraspBall::reset()
{
  /*** グリッパーをフリーの状態に ***/
  auto req_free = std::make_shared<Empty::Request>();
  auto res_free = cli_free_->async_send_request(req_free);
  res_free.wait();
  /*** アームをリセット後３秒間待機 ***/
  std::this_thread::sleep_for(std::chrono::seconds(3));
  /*** システムの状態をリセット ***/
  is_reset_ = true;
}
/*** ボールの三次元情報を受信した時に呼び出されるコールバック関数 ***/
void GraspBall::onPointSubscribed(const PointStamped::SharedPtr point)
{
  if(!is_reset_) return;
  /*** ステレオカメラの座標からマニピュレータの座標へ変換 ***/
  PointStamped tf_point;
  try{
    tf_point = tf_buffer_->transform<PointStamped>(*point, manipulator_name_);
  } catch(tf2::TransformException &ex){
    RCLCPP_ERROR(get_logger(), "%s", ex.what());
    return;
  }
  /*** 物体の位置が閾値外であれば関数を抜ける ***/
  if(tf_point.point.x < x_min_ || tf_point.point.x > x_max_
     || tf_point.point.y < y_min_ || tf_point.point.y > y_max_
     || tf_point.point.z < z_min_ || tf_point.point.z > z_max_) return;
  /*** 別スレッドでアームの制御 ***/
  std::thread{std::bind(&GraspBall::grasp, this, tf_point)}.detach();
  /*** サービスでリセットされるまでアームが動作しないようにする ***/
  is_reset_ = false;
}
/*** リセットする際に呼び出されるコールバック関数 ***/
void GraspBall::serviceReset(const std::shared_ptr<rmw_request_id_t> header,
                             const std::shared_ptr<Empty::Request> req,
                             const std::shared_ptr<Empty::Response> res)
{
  /*** 別スレッドでリセットを実行 ***/
  std::thread{std::bind(&GraspBall::reset, this)}.detach();
}

}
/*** GraspBallクラスをコンポーネントとして登録 ***/
#include <rclcpp_components/register_node_macro.hpp>
RCLCPP_COMPONENTS_REGISTER_NODE(practice_scamper::GraspBall)