#include <rclcpp/rclcpp.hpp>
#include <cv_bridge/cv_bridge.h>
#include <message_filters/subscriber.h>
#include <message_filters/sync_policies/approximate_time.h>
#include <geometry_msgs/msg/point_stamped.hpp>
#include <geometry_msgs/msg/polygon_stamped.hpp>
#include <sensor_msgs/msg/camera_info.hpp>
#include <sensor_msgs/msg/image.hpp>
#include <opencv2/opencv.hpp>

/*** 名前空間を省略して利用できるように宣言 ***/
using geometry_msgs::msg::Point;
using geometry_msgs::msg::Point32;
using geometry_msgs::msg::PointStamped;
using geometry_msgs::msg::PolygonStamped;
using sensor_msgs::msg::CameraInfo;
using sensor_msgs::msg::Image;
using message_filters::Subscriber;
using message_filters::sync_policies::ApproximateTime;
using message_filters::Synchronizer;
/*** typedefを用いた型の再定義 ***/
typedef ApproximateTime<Image, Image, CameraInfo, CameraInfo, PolygonStamped> SyncPolicy;

namespace practice_scamper
{

class Reconstruct3D : public rclcpp::Node
{
  public:
  Reconstruct3D(rclcpp::NodeOptions options = rclcpp::NodeOptions());
  ~Reconstruct3D(){}
  private:
  void initParam();
  void initTopic();
  int calcDisparity(const cv::Mat &img_l, const cv::Mat &img_r, const std::vector<Point32> &points);
  Point reconstruct3D(const CameraInfo &info_l, const CameraInfo &info_r, const std::vector<Point32> &points, int disparity);
  /*** コールバック関数 ***/
  void onImageSubscribed(Image::ConstSharedPtr img_l, Image::ConstSharedPtr img_r,
                         CameraInfo::ConstSharedPtr info_l, CameraInfo::ConstSharedPtr info_r,
                         PolygonStamped::ConstSharedPtr polygon);
  /*** メンバー変数 ***/
  rclcpp::Publisher<PointStamped>::SharedPtr pub_point_;
  std::shared_ptr<Subscriber<Image>> sub_img_l_, sub_img_r_;
  std::shared_ptr<Subscriber<CameraInfo>> sub_info_l_, sub_info_r_;
  std::shared_ptr<Subscriber<PolygonStamped>> sub_polygon_;
  std::shared_ptr<Synchronizer<SyncPolicy>> sync_msgs_;
  /*** パラメータ ***/
  double match_score_;
};
/*** コンストラクタ ***/
Reconstruct3D::Reconstruct3D(rclcpp::NodeOptions options) : Node("reconstruct_3d", options)
{
  initParam();
  initTopic();
}
/*** パラメータの読み込み ***/
void Reconstruct3D::initParam()
{
  match_score_ = declare_parameter<double>("match_score", 0.70);
}
/*** トピックの初期化 ***/
void Reconstruct3D::initTopic()
{
  using namespace std::placeholders;
  /*** パブリッシャの初期化 ***/
  pub_point_ = create_publisher<PointStamped>("point", 10);
  /*** サブスクライバの初期化 ***/
  sub_img_l_.reset(new Subscriber<Image>(this, "/scamper_eyes/left/image"));
  sub_img_r_.reset(new Subscriber<Image>(this, "/scamper_eyes/right/image"));
  sub_info_l_.reset(new Subscriber<CameraInfo>(this, "/scamper_eyes/left/info"));
  sub_info_r_.reset(new Subscriber<CameraInfo>(this, "/scamper_eyes/right/info"));
  sub_polygon_.reset(new Subscriber<PolygonStamped>(this, "rectangle"));
  sync_msgs_.reset(new Synchronizer<SyncPolicy>(SyncPolicy(10),
  *sub_img_l_, *sub_img_r_, *sub_info_l_, *sub_info_r_, *sub_polygon_));
  sync_msgs_->registerCallback(std::bind(&Reconstruct3D::onImageSubscribed, this, _1, _2, _3, _4, _5));
}
/*** ステレオマッチングにより視差を計算 ***/
int Reconstruct3D::calcDisparity(const cv::Mat &img_l, const cv::Mat &img_r, const std::vector<Point32> &points)
{
  cv::Mat temp, input, result;
  cv::Rect temp_rect, input_rect;
  /*** 左右の矩形領域を計算 ***/
  temp_rect = cv::Rect(points[0].y + img_l.cols / 2, img_l.rows / 2 - points[0].x,
  points[2].y - points[0].y, points[0].x - points[2].x);
  input_rect = cv::Rect(0, temp_rect.y, temp_rect.x, temp_rect.height);
  if(temp_rect.width <= 0 || temp_rect.height <= 0) return -1;
  if(input_rect.width <= 0 || input_rect.height <= 0) return -1;
  /*** 左右の画像をグレースケールに変換 ***/
  cv::cvtColor(img_l(temp_rect), temp, cv::COLOR_BGR2GRAY);
  cv::cvtColor(img_r(input_rect), input, cv::COLOR_BGR2GRAY);
  /*** 右画像に対してテンプレートマッチング ***/
  double score;
  cv::Point match_loc;
  cv::matchTemplate(input, temp, result, cv::TM_CCOEFF_NORMED);
  cv::minMaxLoc(result, nullptr, &score, nullptr, &match_loc);
  /*** 視差を計算 ***/
  if(score < match_score_) return -1;
  else return temp_rect.x - match_loc.x;
}
/*** 左右の画像の視差から三次元情報を計算 ***/
Point Reconstruct3D::reconstruct3D(const CameraInfo &info_l, const CameraInfo &info_r, const std::vector<Point32> &points, int disparity)
{
  Point pt_camera, pt_robot;
  /*** カメラパラメータの設定 ***/
  double f = info_l.p[0];
  double baseline = abs(info_r.p[3] / f);
  /*** 物体の三次元座標を計算 ***/
  int img_x = (points[0].x + points[2].x) / 2;
  int img_y = (points[0].y + points[2].y) / 2;
  if(disparity == 0){
    pt_camera.x = std::numeric_limits<double>().max();
    pt_camera.y = pt_camera.z = 0;
  } else{
    pt_camera.z = baseline * f / disparity;
    pt_camera.y = pt_camera.z * img_y / f - baseline / 2;
    pt_camera.x = pt_camera.z * img_x / f;
  }
  /*** カメラ座標からロボット座標への変換 ***/
  pt_robot.x = pt_camera.z;
  pt_robot.y = -pt_camera.y;
  pt_robot.z = pt_camera.x;
  return pt_robot;
}
/*** 画像と矩形領域を受信した時に呼び出されるコールバック関数 ***/
void Reconstruct3D::onImageSubscribed(Image::ConstSharedPtr img_l, Image::ConstSharedPtr img_r,
                                      CameraInfo::ConstSharedPtr info_l, CameraInfo::ConstSharedPtr info_r,
                                      PolygonStamped::ConstSharedPtr polygon)
{
  auto cv_img_l = cv_bridge::toCvShare(img_l);
  auto cv_img_r = cv_bridge::toCvShare(img_r);
  auto &points = polygon->polygon.points;
  /*** ステレオマッチングにより視差を計算 ***/
  int disparity = calcDisparity(cv_img_l->image, cv_img_r->image, points);
  if(disparity < 0) return;
  /*** 左右の画像の視差から三次元情報を計算 ***/
  auto pt_3d = std::make_unique<PointStamped>();
  pt_3d->point = reconstruct3D(*info_l, *info_r, points, disparity);
  pt_3d->header = img_l->header;
  /*** 位置情報をパブリッシュ ***/
  pub_point_->publish(std::move(pt_3d));
}

}
/*** Reconstruct3Dクラスをコンポーネントとして登録 ***/
#include <rclcpp_components/register_node_macro.hpp>
RCLCPP_COMPONENTS_REGISTER_NODE(practice_scamper::Reconstruct3D)