#include <rclcpp/rclcpp.hpp>
#include <rclcpp_action/rclcpp_action.hpp>
#include <geometry_msgs/msg/twist.hpp>
#include <nav_msgs/msg/odometry.hpp>
#include <std_msgs/msg/float64.hpp>
#include <tutorial_interfaces/action/move_robot.hpp>
#include <cmath>

/*** 名前空間を省略して利用できるように宣言 ***/
using geometry_msgs::msg::Pose;
using geometry_msgs::msg::Twist;
using nav_msgs::msg::Odometry;
using std_msgs::msg::Float64;
using MoveRobot = tutorial_interfaces::action::MoveRobot;
using namespace rclcpp_action;
/*** typedefを用いた型の再定義 ***/
typedef std::shared_ptr<ServerGoalHandle<MoveRobot>> GoalHandleMoveRobotPtr;
typedef std::shared_ptr<const MoveRobot::Goal> GoalPtr;

namespace tutorial_scamper
{

class ScamperController : public rclcpp::Node
{
public:
  ScamperController(rclcpp::NodeOptions options = rclcpp::NodeOptions());
  ~ScamperController();
private:
  void execute(const GoalHandleMoveRobotPtr goal_handle);
  /*** コールバック関数 ***/
  void onPoseSubscribed(const Odometry::SharedPtr pose);
  void onDistanceSubscribed(const Float64::SharedPtr distance);
  GoalResponse onGoalSet(const GoalUUID &uuid, GoalPtr goal);
  CancelResponse onActionCanceled(const GoalHandleMoveRobotPtr goal_handle);
  void onActionAccepted(const GoalHandleMoveRobotPtr goal_handle);
  /*** メンバー変数 ***/
  rclcpp::Publisher<Twist>::SharedPtr pub_vel_;
  rclcpp::Subscription<Odometry>::SharedPtr sub_odom_;
  rclcpp::Subscription<Float64>::SharedPtr sub_distance_;
  rclcpp_action::Server<MoveRobot>::SharedPtr action_server_;
  double distance_ = .0;
  double vel_;
  Pose pose_;
};
/*** コンストラクタ ***/
ScamperController::ScamperController(rclcpp::NodeOptions options) : Node("scamper_controller", options)
{
  using std::placeholders::_1;
  using std::placeholders::_2;
  /*** パラメータの読み込み ***/
  vel_ = this->declare_parameter<double>("vel", 0.1);
  /*** パブリッシャの初期化 ***/
  pub_vel_ = this->create_publisher<Twist>("/scamper_driver/cmd_robot_vel", 10);
  /*** サブスクライバの初期化 ***/
  sub_odom_ = this->create_subscription<Odometry>("/scamper_driver/robot_pose", 10,
                                                  std::bind(&ScamperController::onPoseSubscribed, this, _1));
  sub_distance_ = this->create_subscription<Float64>("/scamper_driver/distance", 10,
                                                     std::bind(&ScamperController::onDistanceSubscribed, this, _1));
  /*** アクション(サーバ)の初期化 ***/
  action_server_ = create_server<MoveRobot>(
    this, "move_robot",
    std::bind(&ScamperController::onGoalSet, this, _1, _2),
    std::bind(&ScamperController::onActionCanceled, this, _1),
    std::bind(&ScamperController::onActionAccepted, this, _1)
  );
}
/*** デストラクタ ***/
ScamperController::~ScamperController()
{
}
/*** アクションの実体 ***/
void ScamperController::execute(const GoalHandleMoveRobotPtr goal_handle)
{
  double old_distance = .0;
  Twist stop_vel, target_vel;
  auto goal = goal_handle->get_goal();
  /*** 目標地点の方向からVx, Vyを決定 ***/
  if(goal->target.x == 0 && goal->target.y == 0){
    target_vel = stop_vel;
  } else{
    double target_dir = atan2(goal->target.y, goal->target.x);
    target_vel.linear.x = vel_ * cos(target_dir);
    target_vel.linear.y = vel_ * sin(target_dir);
  }
  /*** 目標地点までの距離を算出 ***/
  double target_distance =
  sqrt(goal->target.x * goal->target.x + goal->target.y * goal->target.y);
  old_distance = distance_;
  /*** ループ周期の設定 ***/
  rclcpp::Rate loop(10);
  /*** フィードバックとリザルトの宣言 ***/
  auto feedback = std::make_shared<MoveRobot::Feedback>();
  auto result = std::make_shared<MoveRobot::Result>();
  /*** 目標地点までの移動制御 ***/
  while(rclcpp::ok()){
    /*** アクションがキャンセルされた場合 ***/
    if(goal_handle->is_canceling()){
      /*** 結果のメッセージを送信 ***/
      result->message = "Canceled";
      goal_handle->canceled(result);
      break;
    }
    /*** 目標地点まで到達した場合 ***/
    else if((distance_ - old_distance) >= target_distance){
      /*** 結果のメッセージを送信 ***/
      result->message = "Succeeded";
      goal_handle->succeed(result);
      pub_vel_->publish(stop_vel);
      break;
    }
    /*** 目標地点まで到達していない場合 ***/
    else{
      /*** 現在の座標を送信 ***/
      feedback->pose = pose_;
      goal_handle->publish_feedback(feedback);
      pub_vel_->publish(target_vel);
    }
    loop.sleep();
  }
}
/*** robot_poseを受信した時に呼び出されるコールバック関数 ***/
void ScamperController::onPoseSubscribed(const Odometry::SharedPtr pose)
{
  pose_ = pose->pose.pose;
}
/*** distanceを受信した時に呼び出されるコールバック関数 ***/
void ScamperController::onDistanceSubscribed(const Float64::SharedPtr distance)
{
  distance_ = distance->data;
}
/*** ゴールが設定された時に呼び出されるコールバック関数 ***/
GoalResponse ScamperController::onGoalSet(const GoalUUID & uuid, GoalPtr goal)
{
  RCLCPP_INFO(this->get_logger(), "Received goal request");
  return GoalResponse::ACCEPT_AND_EXECUTE;
}
/*** アクションがキャンセルされた時に呼び出されるコールバック関数 ***/
CancelResponse ScamperController::onActionCanceled(const GoalHandleMoveRobotPtr goal_handle)
{
  RCLCPP_INFO(this->get_logger(), "Received request to cancel goal");
  return CancelResponse::ACCEPT;
}
/*** アクションが受理された時に呼び出されるコールバック関数 ***/
void ScamperController::onActionAccepted(const GoalHandleMoveRobotPtr goal_handle)
{
  using namespace std::placeholders;
  RCLCPP_INFO(this->get_logger(), "Accepted action");
  std::thread{std::bind(&ScamperController::execute, this, _1), goal_handle}.detach();
}

}
/*** ScamperControllerクラスをコンポーネントとして登録 ***/
#include <rclcpp_components/register_node_macro.hpp>
RCLCPP_COMPONENTS_REGISTER_NODE(tutorial_scamper::ScamperController)