#include <rclcpp/rclcpp.hpp>
#include <rclcpp_action/rclcpp_action.hpp>
#include <std_msgs/msg/float64_multi_array.hpp>
#include <tutorial_interfaces/action/move_robot.hpp>
/*** 名前空間を省略して利用できるように宣言 ***/
using MoveRobot = tutorial_interfaces::action::MoveRobot;
using GoalHandleMoveRobot = rclcpp_action::ClientGoalHandle<MoveRobot>;
using std_msgs::msg::Float64MultiArray;
using namespace rclcpp_action;

namespace tutorial_scamper
{

class ScamperWander : public rclcpp::Node
{
public:
  ScamperWander(rclcpp::NodeOptions options = rclcpp::NodeOptions());
  ~ScamperWander();
private:
  void execute();
  void initParam();
  void initTargetList(std::vector<std::array<float, 2>> &points);
  /*** コールバック関数 ***/
  void onSonarSubscribed(const Float64MultiArray::SharedPtr sonar);
  void onGoalResponseReceived(std::shared_future<GoalHandleMoveRobot::SharedPtr> future);
  void onFeedbackReceived(GoalHandleMoveRobot::SharedPtr goal_handle,
  const std::shared_ptr<const MoveRobot::Feedback> feedback);
  void onResultReceived(const GoalHandleMoveRobot::WrappedResult &result);
  /*** メンバー変数 ***/
  rclcpp::Subscription<Float64MultiArray>::SharedPtr sub_sonar_;
  Client<MoveRobot>::SharedPtr action_client_;
  ResultCode result_code_;
  std::string robot_type_;
  double obstacle_distance_;
  std::vector<double> sonar_;
};
/*** コンストラクタ ***/
ScamperWander::ScamperWander(rclcpp::NodeOptions options) : Node("scamper_wander", options)
{
  using std::placeholders::_1;
  /*** パラメータの読み込み ***/
  initParam();
  /*** サブスクライバの初期化 ***/
  sub_sonar_ = this->create_subscription<Float64MultiArray>("/scamper_sonar/ranges", 10,
  std::bind(&ScamperWander::onSonarSubscribed, this, _1));
  /*** アクション(クライアント)の初期化 ***/
  action_client_ = rclcpp_action::create_client<MoveRobot>(this, "move_robot");
  /*** 別スレッドでアクションを呼び出し ***/
  std::thread{&ScamperWander::execute, this}.detach();
}
/*** デストラクタ ***/
ScamperWander::~ScamperWander()
{
}
/*** 障害物を回避しながらランダム移動を行う関数 ***/
void ScamperWander::execute()
{
  using std::placeholders::_1;
  using std::placeholders::_2;
  std::vector<std::array<float, 2>> target_points{
    std::array<float, 2>{0.25, -0.43}, std::array<float, 2>{0.25, 0.43},
    std::array<float, 2>{-0.5, .0}
  };
  std::array<bool, 3> is_obstacle = {false, false, false};
  bool is_stop = false;
  /*** ループ周期の設定 ***/
  rclcpp::Rate loop(10);
  /*** Scamperをランダム(3方向)に移動 ***/
  auto goal_msg = MoveRobot::Goal();
  while(rclcpp::ok()){
    /*** 乱数をもとに進行方向を決定 ***/
    std::srand(time(NULL));
    unsigned dir = rand() % target_points.size();
    /*** 障害物のある方向には進まないように設定 ***/
    is_stop = false;
    if(is_obstacle[dir]){
      for(int i = 0; i < target_points.size() - 1; i++){
        if(++dir == target_points.size()) dir = 0;
        if(!is_obstacle[dir]) break;
        if(i == target_points.size() - 2) is_stop = true;
      }
    }
    /*** 目標地点の設定 ***/
    if(is_stop){
      goal_msg.target.x = .0f;
      goal_msg.target.y = .0f;
    } else{
      goal_msg.target.x = target_points[dir][0];
      goal_msg.target.y = target_points[dir][1];
    }
    /*** コールバックの登録 ***/
    auto send_goal_options = Client<MoveRobot>::SendGoalOptions();
    send_goal_options.goal_response_callback = std::bind(&ScamperWander::onGoalResponseReceived, this, _1);
    send_goal_options.feedback_callback = std::bind(&ScamperWander::onFeedbackReceived, this, _1, _2);
    send_goal_options.result_callback = std::bind(&ScamperWander::onResultReceived, this, _1);
    /*** アクションの呼び出し ***/
    result_code_ = ResultCode::UNKNOWN;
    auto result = action_client_->async_send_goal(goal_msg, send_goal_options);
    /*** アクションが終了するまで待機 ***/
    while(rclcpp::ok()){
      /*** 障害物のチェック ***/
      for(int i = 0; i < sonar_.size(); i++){
        if(sonar_[i] < obstacle_distance_ && sonar_[i] != .0f) is_obstacle[i] = true;
        else is_obstacle[i] = false;
      }
      /*** 進行方向に障害物が存在すればアクションを中断 ***/
      if(is_obstacle[dir] && !is_stop) {
        action_client_->async_cancel_goal(result.get());
        RCLCPP_INFO(this->get_logger(), "Obstacle was detected");
        while(result_code_ != ResultCode::CANCELED);
        break;
      }
      /*** アクションが完了したらループから抜ける ***/
      if(result_code_ == ResultCode::SUCCEEDED) break;
      loop.sleep();
    }
  }
}
/*** パラメータの初期化を行う関数 ***/
void ScamperWander::initParam()
{
  obstacle_distance_ = this->declare_parameter<double>("obstacle_distance", 0.2);
}
/*** ソナーセンサーの情報を受信したときに呼び出されるコールバック関数 ***/
void ScamperWander::onSonarSubscribed(const Float64MultiArray::SharedPtr sonar)
{
  sonar_ = sonar->data;
}
/*** アクション起動時に呼び出されるコールバック関数 ***/
void ScamperWander::onGoalResponseReceived(std::shared_future<GoalHandleMoveRobot::SharedPtr> future)
{
  auto goal_handle = future.get();
  if (!goal_handle) {
    RCLCPP_ERROR(this->get_logger(), "Goal was rejected by server");
  } else {
    RCLCPP_INFO(this->get_logger(), "Goal accepted by server, waiting for result");
  }
}
/*** フィードバック受信時に呼び出されるコールバック関数 ***/
void ScamperWander::onFeedbackReceived(GoalHandleMoveRobot::SharedPtr goal_handle,
const std::shared_ptr<const MoveRobot::Feedback> feedback)
{
  RCLCPP_INFO(this->get_logger(), "Pose : (%.2f, %.2f)", feedback->pose.position.x, feedback->pose.position.y);
}
/*** アクション終了時に呼び出されるコールバック関数 ***/
void ScamperWander::onResultReceived(const GoalHandleMoveRobot::WrappedResult &result)
{
  result_code_ = result.code;
  switch(result.code){
    /*** アクションが成功した場合 ***/
    case ResultCode::SUCCEEDED:
      RCLCPP_INFO(this->get_logger(), "Result message : %s",
      result.result->message.c_str());
      break;
    /*** アクションが何らかの原因で終了した場合 ***/
    case ResultCode::ABORTED:
      RCLCPP_ERROR(this->get_logger(), "Goal was aborted");
      break;
    /*** アクションが中断された場合 ***/
    case ResultCode::CANCELED:
      RCLCPP_ERROR(this->get_logger(), "Goal was canceled");
      break;
    /*** 上記以外の場合 ***/
    default:
      RCLCPP_ERROR(this->get_logger(), "Unknown result code");
      break;
  }
}
}
/*** ScamperWanderクラスをコンポーネントとして登録 ***/
#include <rclcpp_components/register_node_macro.hpp>
RCLCPP_COMPONENTS_REGISTER_NODE(tutorial_scamper::ScamperWander);